/* 
 * Copyright (C) International Computer Science Institute, 1995/1996.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email bugs to "psather-bugs@icsi.berkeley.edu". 
 *
 * The Lock manager (including dead lock detection)
 *
 * Version 1.0 (released for 1.0.9) by Claudio Fleiner (fleiner@icsi.berkeley.edu)
 * Version 1.1 (released for 1.1  ) by Claudio Fleiner (fleiner@icsi.berkeley.edu)
 */
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>
#include "locks.h"
#include "attach.h"
#include "local.h"
#include "stat.h"
#include "trace.h"
#include "memory.h"
#include <am.h>
#undef DEBUG
#include "debug.h"

#ifdef PSATHER_CHK
#define xDEADLOCK_DETECTION
#endif

int fprintf(FILE *,const char*,...);
int gettimeofday(struct timeval *,void *);
extern int start_sec;
int stat_checks=0;
int stat_moves=0;
int stat_timer=0;

void print_lock_stat()
{
	extern int printf(const char*,...);
	printf("checks: %d\n",stat_checks);
	printf("moves: %d\n",stat_moves);
	printf("timer: %d\n",stat_timer);
}

/*
 * maximum number of outstandig branches
 */
#define MAX_NEW_PLATES 1000
#ifdef DEADLOCK_DETECTION
#define THREAD_GOT_DEADLOCKED (-2)
#endif

/*
 * Maximum number of combinations per Lock
 */
#define MAX_COMBINATIONS 20
/*
 * Maximum number of locks that can be combined
 */
#define MAX_COMBINED_LOCKS 40

#define THREAD_GOT_NO_LOCKS (-1)

/*
 * Each branch of a lock command is encoded in
 * a LOCK_PLATE. The different branches of the same
 * lock statement are kept in single linked list, as
 * are all branches that
 * - are waiting to be served (stored in new_plates)
 * - are waiting for locks (stored in top_plates)
 * - are waiting to be deleted (stored in end_plates)
 * - are currently locked by a thread (stored in locked_plates)
 * - will be moved to the end of the list stored in top_plates 
     (stored in move_plates)
 */

typedef struct LOCK_INFO_struct {
	LOCK lck;
	struct LOCK_INFO_struct *primary;
	struct LOCK_COMB_struct *comb;
	unsigned long seq;
	unsigned long res_seq;
	struct LOCK_PLATE_struct *res_plate;
	thread_t res_tid;
	unsigned hidden_reserved:1;
	unsigned reservable:1;
} LOCK_INFO;

typedef struct LOCK_COMB_struct {
	int lines;
	int cols;
	LOCK_INFO *lck[1];
} LOCK_COMB;
#define NO_COMBINATIONS ((LOCK_COMB*)1)

typedef struct LOCK_PLATE_struct {
	OB_HEADER header; /* needed for the garbage collector ??? 		*/
	struct LOCK_PLATE_struct *next; /* branch in list of all branches 	*/
	struct LOCK_PLATE_struct *next_branch;  /* of the same thread 		*/
	struct LOCK_PLATE_struct *first_branch; /* of the same thread 		*/
	thread_t id;	/* id of the thread requesting those locks 		*/	
	vnn_t reply_to;	/* cluster where the thread is waiting     		*/
	long reply_arg; /* argument used to reply to a threads			*/
	long seq; 	/* is set to the sequence number of the lockmanager loop*/
			/* this way we know if this plate has already been      */
			/* used during a loop					*/
	short cols;	/* number of locks in this plate			*/
	short res;	/* result (-1: NO locks -> execute else branch, other-  */
			/* wise number of branch that was locked		*/
	short branches; /* number of branches in the next_branch list		*/
	short my_branch;/* the number of this branch				*/
	short failed_branches;  /* number of branches that could not be locked  */
	  		        /* with repsect to the branches in next_branch  */
				/* reset for each loop of the lock manager      */
	short flags;
	unsigned short dead;	/* 1 for each lock on this plate that this plate */
				/* did not lock and will never lock		 */
				/* The highest bit is off until the endlock      */
	unsigned short unlocked;   /* 1 for each locak that has been prematurly  */
				   /* unlocked, as we have 16 bits, we cannot    */
				   /* have more than 15 locks per branch         */
				   /* The top bit is reserved in dead.		 */
	struct timeval created,timer_started_at;
		/* time when this plate has been created, and when the last      */
		/* timer has been started					 */
#ifdef DEADLOCK_DETECTION
	short  mark;  /* when doing deadlock detection we mark whether
	  	         or not we've been before */
        short  do_detect_deadlock;
#endif

	LOCK_INFO *lock[1];	/* each lock						 */
} *LOCK_PLATE;

#define ALL_DEAD 0xffff
#define ENDLOCK_DONE 0x8000

struct remote_request_lock {
	thread_t id;
	short flags;
	sema_t lock_arrived;
	short branches;
	short cols;
	LOCK_INFO *lck[1];
};

double l_k=0.001;
static struct LOCK_PLATE_struct top_plate;
static struct LOCK_PLATE_struct new_plates;
static struct LOCK_PLATE_struct end_plates;
static struct LOCK_PLATE_struct locked_plates;
static struct LOCK_PLATE_struct move_plates;
static lock_t lock_new_plates;
static int lock_manager_cluster=0;
static lock_t wakeup_manager; /* the lock manager waits on this lock until there
			       * there is some work to do (new locks, timer went off, .. 
			       */
static lock_t init_lock; /* used to serialize initialization of locks
			  */

static void lock_manager(void);
static void answer_locks(LOCK_PLATE p);
#ifdef DEADLOCK_DETECTION
static int  detect_deadlock(THREAD_ID id);
#endif


/*
 * the lX function are more or less equivalent with 
 * the X function. However, they call the X function so
 * that it executes locally
 */
static void *lEXT_LOCK_primary(LOCK);
static int lEXT_LOCK_reservable(LOCK_PLATE,LOCK_INFO*,unsigned long,THREAD_ID);
static void lEXT_LOCK_reserve(LOCK,THREAD_ID);
static void lEXT_LOCK_free(LOCK,THREAD_ID);
static void lEXT_LOCK_request_reservation(LOCK,THREAD_ID);
static void lEXT_LOCK_cancel_reservation(LOCK,THREAD_ID);
static void *lEXT_LOCK_combinations(LOCK);
static void lEXT_LOCK_combined_with(LOCK,FOB,FOB);
static int lEXT_LOCK_fast_lock(LOCK,THREAD_ID);
static int lEXT_LOCK_fast_try(LOCK,THREAD_ID);
static int lEXT_LOCK_fast_unlock(LOCK,THREAD_ID);

/* if the following macro is defined, all calls on locks
 * are executed locally instead of remotly.
 */
#define USE_LOCAL_EXECUTION

#define NP_LCK_T		lock_t
#define NP_DECL_L_LOCK(a)	lock_t a
#define NP_INIT_L_LOCK(a)	a=LCK_CREATE_LOCKED
#define NP_LOCK_L_LOCK(a)	LCK_LOCK_DELETE(a)
#define NP_LOCK_L_UNLOCK(a) LCK_UNLOCK(a)
#define NP_LOCK_PASS(a)	(long)(a)
#define NP_LCK(a)  ((((NP_VALUE_LOCK*)(a))->lo))
lock_t lck_create_locked() { lock_t t=LCK_CREATE;LCK_LOCK(t);return t; }
#define LCK_CREATE_LOCKED lck_create_locked()
#define LCK_LOCK_DELETE(l) do { LCK_LOCK(l);LCK_DELETE(l); } while(0)

/* making remote calls of the lock manager use polling
 * instead of real locks seams to be a bad idea
 * (performance wise). Therefor it is disabled
 */
#ifdef xxx_POLLING
#define LCK_T		volatile int*
#define DECL_L_LOCK(a)	volatile int a
#define INIT_L_LOCK(a)	a=0
#define LOCK_L_LOCK(a)  { printf("waitfor %p\n",&a);am_wait_for(a);printf("waitfor %p done\n",&a); }
#define LOCK_L_UNLOCK(a)	{ printf("unlocking %p\n",a);*(a)=1; }
#define LOCK_PASS(a)	(long)&(a)
#define LCK(a)  (&(((VALUE_LOCK*)(a))->lo))
#else
#define LCK_T		NP_LCK_T
#define DECL_L_LOCK(a)	NP_DECL_L_LOCK(a)
#define INIT_L_LOCK(a)	NP_INIT_L_LOCK(a)
#define LOCK_L_LOCK(a)	NP_LOCK_L_LOCK(a)
#define LOCK_L_UNLOCK(a) NP_LOCK_L_UNLOCK(a)
#define LOCK_PASS(a)	NP_LOCK_PASS(a)
#define LCK(a)  ((((VALUE_LOCK*)(a))->lo))
#endif

#ifdef DEADLOCK_DETECTION
static void *lEXT_LOCK_wait_for(LOCK,THREAD_ID);
#endif

static int p_reservable_and_visible(LOCK_PLATE p,unsigned long seq);

/*
 * Initialize different lock structures.
 * This function should be executed on ALL
 * clusters.
 */
void p_init_locks()
{
	lock_manager_cluster=0;
	if(AM_MY_PROC==lock_manager_cluster) { 
		lock_new_plates=LCK_CREATE;
		wakeup_manager=LCK_CREATE;
		thr_create_0(AM_MY_PROC,(handler_0_t)lock_manager);
	}
	init_lock=LCK_CREATE;
}

#define addtmv(a,b) 	do { (a).tv_sec+=(b).tv_sec;(a).tv_usec+=(b).tv_usec;if((a).tv_usec>1000000) { (a).tv_usec-=1000000;(a).tv_sec++; } } while(0)
#define subtmv(a,b) 	do { (a).tv_sec-=(b).tv_sec;(a).tv_usec-=(b).tv_usec;if((a).tv_usec<0) { (a).tv_usec+=1000000;(a).tv_sec--; } } while(0)
#define cmptmvi(a,b) 	(((a)>(b))?1:((a)<(b)?-1:0))
#define cmptmv(a,b)  	((a).tv_sec==(b).tv_sec?cmptmvi((a).tv_usec,(b).tv_usec):cmptmvi((a).tv_sec,(b).tv_sec))
#define cptmv(a,b) 	a=b
#define usectmv(a) 	((a).tv_sec*1000000+(a).tv_usec)
#define settmvusec(a,s) do {a.tv_sec=s/100000;a.tv_usec=s%1000000; } while(0)

/*
 * reserve all locks of a lock plate
 * (but don't notify the thread yet)
 */
static void p_lock_up(LOCK_PLATE p)
{
	int i;
	p->dead=ALL_DEAD-ENDLOCK_DONE;
	p->unlocked=ALL_DEAD-ENDLOCK_DONE;
	for(i=0;i<p->cols;i++) {
		p->dead&= ~(1<<i);
		p->unlocked&= ~(1<<i);
		lEXT_LOCK_reserve(p->lock[i]->lck,p->id);
		if(p->flags&LOCK_HIDDEN) p->lock[i]->primary->hidden_reserved=1;
		else p->lock[i]->primary->hidden_reserved=0;
	}
	p->flags|=PLATE_IS_LOCKED;
}

/*
 * Undo a previous p_lock_up (used only in x_find_combinations)
 */
static void p_unlock(LOCK_PLATE p)
{
	int i;
	p->dead=0;
	p->unlocked=0;
	for(i=0;i<p->cols;i++) {
		lEXT_LOCK_free(p->lock[i]->lck,p->id);
		p->lock[i]->primary->hidden_reserved=0;
	}
	p->flags&=~PLATE_IS_LOCKED;
}
 
/*
 * Removes all branches from top_plates that have been
 * generated by the thread that owns a. Actually, the plates
 * are not removed, but simply marked as dead.
 */
static void remove_remaining_branches(LOCK_PLATE a)
{
	LOCK_PLATE t;
	int i;
	t=a->first_branch;
	while(t!=NULL) {
		if(t!=a) {
			DEBUG1("killing branch %p",t);
			t->dead=ALL_DEAD;
			t->res=THREAD_GOT_NO_LOCKS;
		}
		for(i=0;i<t->cols;i++)
			lEXT_LOCK_cancel_reservation(t->lock[i]->lck,t->id);
		t=t->next_branch;
	}
}

/*
 * Checks the combination rules.
 * x_combination_is_ok is the recursive function started by combination_is_ok.
 */
static int combination_done(short *used,int size)
{
	while(--size>=0) if(!used[size]) return 0;
	return 1;
}

static int applicable(LOCK_INFO *rule[],int rule_size,int current,LOCK_INFO *comb[],short *used,int size,int recursion)
{
	int r,i,j;

#ifdef LOCK_DEBUG
	int k;
	printf("%d. testing rule ",recursion);
	for(k=0;k<rule_size;k++) 
		if(rule[k]!=NULL) 
			printf("%p ",rule[k]);
	printf("\n");
#endif

	for(i=0;i<rule_size;i++) {
		if(rule[i]==NULL) return 0; /* comb[current] is not part of this rule */
		if(comb[current]==rule[i]) break;
	}
	if(i==rule_size) return 0; /* comb[current] is not part of this rule */
	r=i;
	used[current]=recursion;
	for(i=0;i<rule_size && rule[i]!=NULL;i++) {
		if(r!=i) {
			for(j=0;j<size;j++) {
				if(comb[j]==rule[i] && used[j]==0) {
					used[j]=recursion;
					break;
				}
			}
			if(j==size) {
				for(j=0;j<size;j++) 
					if(used[j]==recursion) used[j]=0;
				return 0;
			}
		}
	}
#ifdef LOCK_DEBUG
	printf("yes, it works\n");
#endif
	return 1;
}
static int x_combination_is_ok(LOCK_INFO *comb[],short *used,int size,int recursion)
{
	int i,j,cl;
	LOCK_COMB *c;
	/* find the first lock that is not already part of a combination */
	for(i=0;i<size;i++) {
		if(used[i]==0) {
			c=comb[i]->primary->comb;
			for(cl=0;cl<c->lines;cl++) {
				if(applicable(c->lck+cl*c->cols,c->cols,i,comb,used,size,recursion)) {
#ifdef LOCK_DEBUG
					int k;
					printf("Result: ");
					for(k=0;k<size;k++)
						printf("%3d ",used[k]);
					printf("\n");
#endif

					if(combination_done(used,size) ||
					   x_combination_is_ok(comb,used,size,recursion+1)) return 1;
					for(j=i;j<size;j++)
						if(used[j]==recursion) used[j]=0;
				}
			}
		}
	}
	return 0;
}

static short used[MAX_COMBINED_LOCKS];
static short used_size;
static int combination_is_ok(LOCK_INFO *comb[],int size)
{
	int i;
#ifdef LOCK_DEBUG
	LCK_LOCK(lock_new_plates);
	printf("testing the follwoing combination:\n");
	for(i=0;i<size;i++) 
		printf("%p ",comb[i]);
	printf("\n");
#endif
	used_size=size;
	memset(used,0,sizeof(short)*MAX_COMBINED_LOCKS);
	i=x_combination_is_ok(comb,used,size,1);
#ifdef LOCK_DEBUG
	if(i) printf("OKOKOKOKOK\n");
	else printf("we need to search for more plates\n");
	LCK_UNLOCK(lock_new_plates);
#endif
	return i;
}
/*
 * Check if the plate (or another one of the same thread)
 * is already in the combination)
 */
static int already_in_combination(LOCK_PLATE p) 
{
	p=p->first_branch;
	while(p!=NULL) {
		if(p->flags&PLATE_IN_COMBINATION) return 1;
		p=p->next_branch;
	}
	return 0;
}
/*
 * We know that we can lock p, but we also know that we need
 * some other plates as well, as some of the locks of p need
 * combinations. x_find_combinations is the actual function,
 * find_combinations is just used to start the recursion.
 * We create all possible permutations of plates and check each one
 * if it satisfies the combination rules.
 */
static int x_find_combinations(LOCK_PLATE p,int seq,LOCK_INFO *comb[],thread_t comb_tids[],int size)
{
	int i,new_size;
	/* first we lock p, and add all locks that need combinations to comb */
	while(p!=NULL) {
		if(!already_in_combination(p) && p->flags&PLATE_USES_COMBINATIONS 
		   && !p->dead && p_reservable_and_visible(p,seq)) {
			p_lock_up(p);
			new_size=size;
			for(i=0;i<p->cols;i++) 
				if(p->lock[i]->primary->comb!=NO_COMBINATIONS) {
					comb_tids[new_size]=p->id;
					comb[new_size++]=p->lock[i];
				}
			assert(new_size!=size);
			p->flags|=PLATE_IN_COMBINATION;
			if(combination_is_ok(comb,new_size)?1:
			   x_find_combinations(p->next,seq,comb,comb_tids,new_size)) {
				remove_remaining_branches(p);
				return 1;
			}
			p->flags&=~PLATE_IN_COMBINATION;
			p_unlock(p);
		}
		p=p->next;
	}
	return 0;
}

static int find_combinations(LOCK_PLATE p,int seq)
{
	int i,j,k,size;
	LOCK_INFO *comb[MAX_COMBINED_LOCKS];
	thread_t comb_tids[MAX_COMBINED_LOCKS];
	struct {
		OB_HEADER header;
		int asize;
		FOB arr_part[MAX_COMBINED_LOCKS];
	} *lcks;
	struct {
		OB_HEADER header;
		int asize;
		thread_t arr_part[MAX_COMBINED_LOCKS];
	} *tids;

	p_lock_up(p);

	for(i=size=0;i<p->cols;i++) {
		if(p->lock[i]->primary->comb!=NO_COMBINATIONS) {
			comb_tids[size]=p->id;
			comb[size++]=p->lock[i];
		}
	}
	p->flags|=PLATE_IN_COMBINATION;
	i=combination_is_ok(comb,size)?1:x_find_combinations(p->next,seq,comb,comb_tids,size);
	p->flags&=~PLATE_IN_COMBINATION;
	if(!i) {
		p_unlock(p);
		return(0);
	}
	/* we have to call EXT_LOCK_combined_with for all
	 * combinations. The array 'used' defines which locks
	 * are to be locked together.
	 */
	MALLOC(tids,sizeof(*tids));
	MALLOC(lcks,sizeof(*lcks));
	for(i=1;i<=used_size;i++) {
		FOB prim=NULL;
		for(k=j=0;j<used_size;j++) {
			if(used[j]==i) {
				tids->arr_part[k]=comb_tids[j];
				lcks->arr_part[k++]=comb[j]->lck;
				prim=comb[j]->primary->lck;
			}
		}
		if(k==0) break;
		tids->asize=lcks->asize=k;
		lEXT_LOCK_combined_with(prim,lcks,tids);
	}
	FREE(lcks);
	FREE(tids);
	return(1);
}


/*
 * checks if all locks of a plate are reservable and that no
 * other plate waits for one of those locks. The second check
 * is done as follows: 
 * If a lock plate does not allow another one to reserve (lock) a
 * lock, it marks it as rain checked. (see p_mark_locks_as_rain_checked).
 * This is done by storing the current lock manager sequence number
 * in the lock. If this sequence number is equal to the current sequence
 * number seq, the lock is unavailable
 * If it can get all locks, it will also try to satisfy combination
 * locks by locking at the next plates. This is a recursive function
 * with back tracking, and hence not too efficient.
 */
static int p_reservable_and_visible(LOCK_PLATE p,unsigned long seq)
{
	int i;
#ifdef LOCK_TRACE
	char pp[30];
#endif
	for(i=0;i<p->cols;i++) {
		if(!lEXT_LOCK_reservable(p,p->lock[i],seq,p->id)) {
			LOCK_TR2("for %s: reservable for [%ld] failed",thr_print_id(p->id,pp),(long)p->lock[i].lck);
			return 0;
		}
		if(p->lock[i]->primary->seq==seq) {
			LOCK_TR2("for %s: check for [%ld] ok, but end invisible",thr_print_id(p->id,pp),(long)p->lock[i].lck);
			return 0;
		}
	}
	LOCK_TR2("for %s: reservable for [%ld] succeeded",thr_print_id(p->id,pp),(long)p->lock[i].lck);
	return 1;
}
static int p_reservable_and_visible_then_lock(LOCK_PLATE p,unsigned long seq)
{
	if(!p_reservable_and_visible(p,seq)) return 0;
	/* Do we need to check for combinations ?*/
	if(!(p->flags&PLATE_USES_COMBINATIONS)) {
		p_lock_up(p);
		return 1;
	}
	return(find_combinations(p,seq));
}


/*
 * A lock marked as rain checked cannot be locked during
 * the current loop of the lock manager.
 * This is done whenever the current plate has waited long enough 
 * to have a higher precedence for this lock than any other plate
 * waiting for it.
 */
static void p_mark_locks_as_rain_checked(LOCK_PLATE p,unsigned long seq)
{
	int i;
	for(i=0;i<p->cols;i++)
		if(lEXT_LOCK_reservable(p,p->lock[i],seq,p->id)) {
			LOCK_TR1("marking [%d] as rain_checked",p->lock[i].lck);
			p->lock[i]->primary->seq=seq;
		}
}

/*
 * Checks to see if no lock for this plate is
 * actually reservable
 */
static int p_nothing_is_reservable(LOCK_PLATE p,unsigned long seq)
{
	int i;
	for(i=0;i<p->cols;i++) {
		if(lEXT_LOCK_reservable(p,p->lock[i],seq,p->id)) return 0;
	}
	return 1;
}

/*
 * checks to see if all locks for a current plate are reservable 
 */
static int p_cond_ok(LOCK_PLATE p,unsigned long seq)
{
	int i;
	for(i=0;i<p->cols;i++)
		if(!lEXT_LOCK_reservable(p,p->lock[i],seq,p->id)) return 0;
	return 1;
}

/*
 * calculates the time when this lock plate either must have
 * all its locks, or else it will be moved to the end of the
 * list of plates waiting for locks
 */
static void p_get_time_to_move(LOCK_PLATE p,struct timeval *time_move)
{
	struct timeval now,t;
	long age;
	gettimeofday(&now,NULL);
	subtmv(now,p->created);
	age=usectmv(now);
	age=((l_k*(double)age));
	*time_move=p->timer_started_at;
	settmvusec(t,age);
	addtmv(*time_move,t);
}

/* 
 * returns true if one of the locks is reserved (locked),
 * but ignores all hidden reserved locks (a lock is hidden
 * reserved, if we want to avoid that a lock executes an else
 * part even though one of the locks it needs is actually 
 * locked. This is used for example during gate enqueues and
 * dequeues. Enqueues and dequeues are defined to be invisible,
 * even though they lock the gate momentarly
 */
static int one_locked_or_cond_wrong_but_unreserved(LOCK_PLATE p,long seq)
{
	int i;
	for(i=0;i<p->cols;i++) {
		if(p->lock[i]->primary->hidden_reserved) continue;
		if(!lEXT_LOCK_reservable(p,p->lock[i],seq,p->id)) return 1;
	}
	return 0;
}

/*
 * delete all end plates, iff all the locks
 * that have been locked because of them have been unlocked
 * again.
 */
static void process_end_plates()
{
	LOCK_PLATE act,t1;
	int i,dead;

	act= &end_plates;
	while(act->next!=NULL) {
		t1=act->next;
		/* t1->dead could change due to an endlock, 
		 * so we store the current value
		 * locally, instead of using a lock
		 */
		if((dead=t1->dead)) {
			if(t1->res!=THREAD_GOT_NO_LOCKS) 
				for(i=0;i<t1->cols;i++)
					if((dead&(1<<i)) && !(t1->unlocked&(1<<i))) {
						lEXT_LOCK_free(t1->lock[i]->lck,t1->id);
						t1->unlocked|=1<<i;
					}

			if(dead==ALL_DEAD) {
				act->next=t1->next;
				FREE(t1);
			} else act=act->next;
		} else act=act->next;
	}
}

/*
 * Read the combination table into local structures
 */
static void read_combinations(LOCK_INFO *info)
{
	struct L_A_s {
		OB_HEADER header;
		int asize;
		struct L_A_s *arr_part[1];
	} *rem_array;
	struct L_A_s *columns[MAX_COMBINATIONS];
	int colsize[MAX_COMBINATIONS];
	int lines=0,cols=0;
	int pos,i;
	LOCK_COMB *c;
	sema_t sem;

	sem=SEMA_CREATE(0);
	rem_array=lEXT_LOCK_combinations(info->lck);
	if(FVOID(rem_array)) {
		info->comb=NO_COMBINATIONS;
		return;
	}
	pos=WHERE(rem_array);
	lines=F_ASIZE(struct L_A_s *,rem_array);
	if(lines>MAX_COMBINATIONS) {
		fprintf(stderr,"WARNING: I cannot handle more than 20 combinations, ignoring rest\n");
		lines=MAX_COMBINATIONS;
	}

	am_get(pos,SENDFOBHOME(rem_array->arr_part),columns,lines*sizeof(FOB),r_sema_signal_mem,sem);
	SEMA_WAIT(sem);
	for(i=0;i<lines;i++) {
		columns[i]=RECVFOB(columns[i],pos);
		colsize[i]=F_ASIZE(struct L_A_s *,columns[i]);
		if(cols<colsize[i]) cols=colsize[i];
	}
	CALLOC(c,cols*lines*sizeof(void *)+sizeof(*c),1);
	c->lines=lines;
	c->cols=cols;
	for(i=0;i<lines;i++) {
		pos=WHERE(columns[i]);
		am_get(pos,SENDFOBHOME(columns[i]->arr_part),c->lck+i*cols,colsize[i]*sizeof(FOB),r_sema_signal_mem,sem);
	}
	for(i=0;i<lines;i++) SEMA_WAIT(sem);
	for(i=0;i<lines*cols;i++)
		if(!FVOID(c->lck[i]))
			F_R_RATTR_NA(c->lck[i],LOCK,c->lck[i],LOCK_header);
		
	info->comb=c;
	
#ifdef LOCK_DEBUG
	LCK_LOCK(lock_new_plates);
	printf("GOT COMBINATION FOR LOCK %p info=%p",info->lck,info);
	for(i=0;i<c->lines;i++) {
		printf("\n%d. ",i);
		for(pos=0;pos<c->cols;pos++)
			if(c->lck[i*c->lines+pos]) printf("%p ",c->lck[i*c->lines+pos]);
	}
	printf("\n---------------------------------\n");
	LCK_UNLOCK(lock_new_plates);
#endif
}
	
/* 
 * Read the primary for locks where it is not yet
 * known. 
 */
static void init_plates(LOCK_PLATE np)
{
	int i;
	LOCK_PLATE p=np;

	while(p!=NULL) {
#ifdef DEADLOCK_DETECTION
		p->do_detect_deadlock = 0;
#endif
		for(i=0;i<p->cols;i++) {
			if(p->lock[i]->primary==NULL) {
				LOCK prlck;
				prlck=(LOCK)lEXT_LOCK_primary(p->lock[i]->lck);
				F_R_RATTR_NA(p->lock[i]->primary,LOCK,prlck,LOCK_header);
				if(p->lock[i]->primary->comb==NULL)
					read_combinations(p->lock[i]->primary);
			}
			assert(p->lock[i]->primary!=NULL);
			if(p->lock[i]->primary->comb!=NO_COMBINATIONS) {
				p->flags|=PLATE_USES_COMBINATIONS;
			}
			lEXT_LOCK_request_reservation(p->lock[i]->lck,p->id);

		}
		p=p->next;
	}
}

#ifdef LOCK_DEBUG
static void chkplate()
{
	FILE *lff=NULL;
	LOCK_PLATE t;
	lff=fopen("lock_plates","a");
	t=top_plate.next;
	while(t!=NULL)  {
		fprintf(lff,"0x%p  ->  0x%p\n",t,t->first_branch);
		t=t->next;
	}
	fprintf(lff,"------------------------------\n");
	fflush(lff);
	fclose(lff);
}
#endif

static void awake_lock_manager(vnn_t from) 
{
	LCK_UNLOCK(wakeup_manager);
}

/*
 * Finally,the lock manager thread function. This
 * function is executed by a separate thread on exactly
 * one cluster.
 */
static void lock_manager()
{
	LOCK_PLATE act,t1=NULL,np;
	volatile unsigned long seq=0;
	int print_lock_info=0;
	struct timeval now,next_timer,time_move,min_timer_int;
	int need_timer,need_deadlock_timer;
	extern void deamon_thread();
	min_timer_int.tv_sec=0;
	min_timer_int.tv_usec=5000;
	init_thread();
	deamon_thread();
	((LOCAL_MEM)THR_LOCAL)->file="LOCK_MANAGER";
	((LOCAL_MEM)THR_LOCAL)->line=0;

	INIT_PROTECT_BEGIN;
	assert(sizeof(THREAD_ID)==8);
#ifdef SOLARIS_THREADS
	if(getenv("LOCK_MANAGER_PRIO")) {
		int p;
		assert(thr_setprio(thr_self(),atoi(getenv("LOCK_MANAGER_PRIO")))==0);
#ifdef POLLING
		if(p<=2) p=2;
#endif
		thr_getprio(thr_self(),&p);
		printf("lock manager priority set to %d\n",p);
	}
#endif
	if(getenv("PRINT_LOCK_INFO")) print_lock_info=1;

	while(1) {
		/* wait until there is some work (new plates, timer went off) */
		LCK_LOCK(wakeup_manager);

		if(print_lock_info) printf("lock manager runs another round\n");
		LOCK_TR1("LOCK MANAGER STARTS ANOTHER ROUND (%d)",seq);

		seq++; /* the sequence number, used to 'rain check' locks */
		gettimeofday(&now,NULL);
		next_timer.tv_sec=0x7fffffff;
		next_timer.tv_usec=0;
		need_timer=0;



		/* add all new plates to the central list */
		np=new_plates.next;
		if(np!=NULL) { 
			LCK_LOCK(lock_new_plates);
			np=new_plates.next;
			new_plates.next=NULL;
			LCK_UNLOCK(lock_new_plates);
			SYS_IMPORT;
			init_plates(np);
			act=&top_plate;
			while(act->next!=NULL) act=act->next;
			act->next=np;
		} else SYS_IMPORT;
#ifdef LOCK_DEBUG
		chkplate();
#endif
		/* start by deleting obsolete plates 
		 * We have to do this here, after we got the new plates,
		 * because we have to make sure that there are no new plates
		 * comming. Otherwise, a thread could end a lock and mark
		 * the plate as dead after process_end_plates(), and add
		 * a request for new plates before the lock manager had
		 * a chance to delete the old lock plate.
		 */
		process_end_plates();

		/* check if any of the plates could get its lock */
		act= &top_plate;
		while(act->next) {
			t1=act->next;
			/* if we have a "living" plate (not marked as dead by 
			 * remove_remaining_branches()), we set the sequence number
			 * of the first plate for this thread to the current sequence number
			 * and initialize a counter used to find out if all plates are
			 * stuck in top_plates (in this case we would execute the 
			 * else part (if it exists at all)
			 */
			if(t1->dead==0 && t1->first_branch->seq!=seq) {
				t1->first_branch->seq=seq;
				t1->first_branch->failed_branches=0;
			}

			stat_checks++;  /* some statistics */
#ifdef DEADLOCK_DETECTION
                        DEBUG0("might do a deadlock detect");
                        if (t1->do_detect_deadlock) {
                            t1->do_detect_deadlock = 0;
                            if (detect_deadlock(t1->first_branch->id)) {
                                /* yes, we are in a deadlock.  Now we need */
                                /* to send a signal back to the thread. */
                                fprintf(stderr, "DEADLOCK DETECTED\n");
                                act->next=t1->next;
                                t1->next=locked_plates.next;
                                t1->res=THREAD_GOT_DEADLOCKED;
                                locked_plates.next=t1;
                                remove_remaining_branches(t1);
				answer_locks(t1);
                                break;
                            }
                        }
#endif
			/* move dead plates to the end_plates list. We cannot
			 * do that in remove_remaining_branches, because it is
			 * quite expensive to delete an element in a single linked
			 * list if the predecessor is not known. Here we know it,
			 * hence we move the plate here.
			 */
			if(t1->dead==ALL_DEAD) {
				act->next=t1->next;
				t1->next=end_plates.next;
				end_plates.next=t1;

			/* A plate can already be locked (because of
			 * RENDEZVOUS types of locks. In this case just move
			 * it.
			 */
			} else if(t1->flags&PLATE_IS_LOCKED) {
				act->next=t1->next;
				t1->next=locked_plates.next;
				t1->res=t1->my_branch;
				locked_plates.next=t1;
				answer_locks(t1);

			/* Well, if the plate can get all its lock,
			 * lock'em up and remove all other branches
			 * of the same thread
			 */
			} else if(p_reservable_and_visible_then_lock(t1,seq)) {
				act->next=t1->next;
				t1->next=locked_plates.next;
				t1->res=t1->my_branch;
				locked_plates.next=t1;
				remove_remaining_branches(t1);
				answer_locks(t1);

			/* If there is an else part, and all locks for this
			 * plate are reserved (and none of them as hidden),
			 * increment the failed_branches of the first plate
			 * (=branch) of this thread, and if all branches have failed
			 * we notify the thread that it should proceed with the else
			 * branch.
			 */
			} else if(t1->flags&LOCK_WITH_ELSE && one_locked_or_cond_wrong_but_unreserved(t1,seq)) {
				t1->first_branch->failed_branches++;
				if(t1->first_branch->failed_branches==t1->branches) {
					act->next=t1->next;
					t1->next=locked_plates.next;
					t1->res=THREAD_GOT_NO_LOCKS;
					locked_plates.next=t1;
					remove_remaining_branches(t1);
					answer_locks(t1);
				} else act=act->next;

			/* if all locks are reserved, ignore this plate
			 * (there is no point in moving another plate, as this
			 * would have no effect for this plate)
			 */
			} else if(p_nothing_is_reservable(t1,seq)) {
#ifdef DEADLOCK_DETECTION
				if (!t1->first_branch->do_detect_deadlock) {
					DEBUG0("need timer for nothing_is_reservable");
					/* we use another timer for deadlock detection, as
					   the standard one would resume much to early
					   and result in a very busy lock manager */
					need_deadlock_timer=1; 
					t1->first_branch->do_detect_deadlock = 1;
				}
#endif
				act=act->next;

			/* However, if all locks are actually free, but
			 * blocked by other plates that have a higher priority, 
			 * we set the timer flag to ensure that those plates
			 * eventually move away
			 */
			} else if(p_cond_ok(t1,seq)) {
				DEBUG0("need timer");
#ifdef DEADLOCK_DETECTION
                                t1->first_branch->do_detect_deadlock = 1;
#endif
				need_timer=1;
				act=act->next;

			/* This plate could get some of the locks,
			 * but not all. If the timer for this plate already
			 * expired, move it awy. Otherwise mark the locks as
			 * 'rain checked' and calculate the time when the
			 * timer should go off
			 */
			} else {
				p_get_time_to_move(t1,&time_move);
				subtmv(time_move,min_timer_int);
				if(cmptmv(now,time_move)>=0) {
					gettimeofday(&t1->timer_started_at,NULL);
					act->next=t1->next;
					t1->next= move_plates.next;
					move_plates.next=t1;
					DEBUG0("plate moved");
					stat_moves++;
				} else {
					addtmv(time_move,min_timer_int);
					if(cmptmv(next_timer,time_move)>=0)
						cptmv(next_timer,time_move);
					p_mark_locks_as_rain_checked(t1,seq);
					LOCK_TR2("timer set to %ld.%06ld",next_timer.tv_sec,next_timer.tv_usec);
					act=act->next;
				}
			}
		}
		act->next=move_plates.next;
		move_plates.next=NULL;


		/* answer all threads that got their locks
		 * Actually we could do that above, instead of creating a new
		 * list. But as the lock manager holds the lock lock_plate,
		 * and this lock is also used in some reply / request functions,
		 * the lock_manager is not allowed to send out requests as long
		 * as it holds this lock (lock_plate)
		 */
		SYS_EXPORT; /* make sure that all those threads get the correct values */
		while(locked_plates.next!=NULL) {
			t1=locked_plates.next;
			locked_plates.next=t1->next;
			t1->next=end_plates.next;
			end_plates.next=t1;
			if(t1->res==THREAD_GOT_NO_LOCKS) {
				DEBUG1("killing plate %p",t1);
				t1->dead=ALL_DEAD;
			}
			/* answer_locks(t1); */
		}
		if(need_timer || need_deadlock_timer) {
#ifdef DEADLOCK_DETECTION
			if(!need_timer && need_deadlock_timer) {
				gettimeofday(&next_timer,NULL);
				next_timer.tv_sec+=5; /* check again in 5 seconds */
			}
			need_deadlock_timer=0;
#endif
			if(next_timer.tv_sec!=0x7fffffff) {
				delay_t d;
				gettimeofday(&now,NULL);
				if(cmptmv(now,next_timer)>0) {
					LCK_UNLOCK(wakeup_manager);
				} else {
					LOCK_TR2("next_timer=%d.%06d",next_timer.tv_sec,next_timer.tv_usec);
					subtmv(next_timer,now);
					LOCK_TR2("now=%d.%06d",now.tv_sec,now.tv_usec);
					LOCK_TR2("next_timer-now=%d.%06d",next_timer.tv_sec,next_timer.tv_usec);
					d.sec=next_timer.tv_sec;
					d.nsec=1000*next_timer.tv_usec;
					LOCK_TR2("timer started, should ring in %d.%09d",d.sec,d.nsec);
					stat_timer++;
					thr_delay_function(d,(handler_delay_t)lck_unlock,wakeup_manager);
				}
			} else {
				LCK_UNLOCK(wakeup_manager); /* simulate a zero wait timer return */
			}
		}
	}

	INIT_PROTECT_END;
}

/*
 * called on the lock_manager_cluster, creates a new plate and stores it in 
 * new_plates. Does not reply, unless it had to create a new LOCK_INFO 
 * structure.
 */
static void store_lock(vnn_t from,struct remote_request_lock *rrl,int size,long rl)
{
	LOCK_PLATE last=NULL,top=NULL,n;
	int i,j,k;

	LCK_LOCK(lock_new_plates);
	j=(rand()>>3)%rrl->branches;
#ifdef LOCK_DEBUG
	printf("STORE_LOCK:\n");
#endif
	for(i=0;i<rrl->branches;i++,j=(j+1)%rrl->branches) {
#ifdef LOCK_DEBUG
		printf("when %d (%d): ",j,i);
#endif
		if(rrl->lck[j*rrl->cols]==NULL) continue;
		CALLOC(n,sizeof(*n)+sizeof(DLOCK)*(rrl->cols),1);
		n->id=rrl->id;
		n->reply_to=from;
		n->reply_arg=rl;
		n->branches=rrl->branches;
		n->cols=rrl->cols;
		n->my_branch=j;
		n->flags=rrl->flags;
		if(top==NULL) top=n;
		n->res=THREAD_GOT_NO_LOCKS;
		n->first_branch=top;
		n->next_branch=NULL;
		if(last!=NULL) last->next_branch=n;
		last=n;
		for(k=0;k<n->cols;k++) {
			LOCK_INFO *x=rrl->lck[k+j*rrl->cols];
			if(x==NULL) n->cols=k;
			else {
				n->lock[k]= rrl->lck[k+j*rrl->cols];
#ifdef LOCK_DEBUG
				printf("%p ",n->lock[k]);
#endif
			}
		}
#ifdef LOCK_DEBUG
		printf("\n");
#endif

		gettimeofday(&n->created,NULL);
		gettimeofday(&n->timer_started_at,NULL);
		n->next=new_plates.next;
		new_plates.next=n;
	}
#ifdef LOCK_DEBUG
	printf("\n--------------------------------\n");
#endif
	LCK_UNLOCK(wakeup_manager);
	LCK_UNLOCK(lock_new_plates);
	FREE(rrl);
}

/*
 * executed on the same node as the thread that requested the lock.
 * executed by an am_request from the lock_manager_thread
 */
static void reply_lock(vnn_t from,struct request_lock *rl,long plate,short res)
{
	rl->plate=plate;
	rl->res=res;
	DEBUG3("rl=%p  res=%d   sem=%p",rl,rl->res,rl->sem);
	LCK_UNLOCK(rl->lck);
}

/*
 * executed by the lock_manager to answer the request for locks
 */
static void answer_locks(LOCK_PLATE p)
{
	assert(p->res>=-2);
	am_request_3(p->reply_to,(handler_3_t)reply_lock,p->reply_arg,(long)p,p->res);
}

#ifdef LOCK_TRACE
static char *print_locks(struct EXCEPT_ELEMENT0 *ex,int res)
{
	int i,j;
	char *p;
	assert(ex->branches>0 && ex->cols>0);
	CALLOC(p,ex->cols*ex->branches*10+20,1);
	p[0]=0;
	if(ex->flags&LOCK_HIDDEN) strcat(p,"R:");
	if(ex->flags&LOCK_WITH_ELSE) strcat(p,"ELSE:");
	strcat(p,"[");
	for(j=0;j<ex->branches;j++) {
		if(j!=0) strcat(p,"| ");
		if(j==res) strcat(p,"=> ");
		for(i=0;i<ex->cols;i++)
			sprintf(p+strlen(p),"%p ",ex->locks[i+j*ex->cols]);
		if(j==res) strcat(p,"<= ");
	}
	p[strlen(p)-1]=']';
	return p;
}
#endif

int r_lock(void *vex)
{
	EXCEPT_LOCK_ELEMENT(1) *ex=vex;
	struct remote_request_lock *rrl;
	void *rem;
	struct request_lock rl;
	int i;
#ifdef LOCK_TRACE
	char *p;
#endif

	STAT(LOCK_STMT);
	if(ex->branches*ex->cols==1) {
		if(ex->flags&LOCK_WITH_ELSE) {
			switch(lEXT_LOCK_fast_try(ex->locks[0],thr_ps_id())) {
			case -1: /* use else part */
				STAT(FAST_LOCK);
				SYS_IMPORT;
				return -1;
			case 1:  /* got lock */
				STAT(FAST_LOCK);
				ex->flags|=LOCK_LOCAL;
				if(!(ex->flags&LOCK_NO_STACK)) p_ex_push(ex);
				SYS_IMPORT;
				return 0;
			} /* call lock manager */
		} else if(lEXT_LOCK_fast_lock(ex->locks[0],thr_ps_id())) {
			STAT(FAST_LOCK);
			ex->flags|=LOCK_LOCAL;
			if(!(ex->flags&LOCK_NO_STACK)) p_ex_push(ex);
			SYS_IMPORT;
			return 0;
		}
	}
	CALLOC(rrl,sizeof(*rrl)+sizeof(LOCK_INFO *)*ex->branches*ex->cols,1);
	rem=r_alloc(lock_manager_cluster,sizeof(*rrl)+sizeof(DLOCK *)*ex->branches*ex->cols);

	for(i=0;i<ex->branches*ex->cols;i++) {
		FOB f;
		if(!FVOID(ex->locks[i])) { 
			F_R_RATTR_NA(f,LOCK,ex->locks[i],LOCK_header);
			rrl->lck[i]=SENDFOBHOME(f);
		} else {
			rrl->lck[i]=NULL;
		}
	}
	rrl->flags=ex->flags;
	rrl->id=thr_ps_id();
	rrl->branches=ex->branches;
	rrl->cols=ex->cols;
	rl.lck=LCK_CREATE;LCK_LOCK(rl.lck);
	rl.res=-3;
	rl.plate= -3;

#ifdef LOCK_TRACE
	p=print_locks(ex,-1);
	LOCK_TR1("( lock %s",p);
	FREE(p);
#endif

	am_store(lock_manager_cluster,
		rrl,SENDFOBHOME(rem),sizeof(*rrl)+sizeof(DLOCK *)*ex->branches*ex->cols,
		(handler_mem_t)store_lock,&rl);

	REG(lck_unlock,rl.lck);
	LCK_LOCK(rl.lck);
	UNREG;
	LCK_DELETE(rl.lck);

	ex->plate=rl.plate;
	DEBUG3("rl=%p  rl.res=%d   rl.sem=%p",&rl,rl.res,rl.sem);
	assert(rl.res>=-2);
	if(rl.res>=0) {
		assert(rl.plate!=0);
		if(!(ex->flags&LOCK_NO_STACK)) p_ex_push(ex);
	}

#ifdef LOCK_TRACE
	p=print_locks(ex,rl.res);
	if(rl.res>=0) LOCK_TR1(") lock succeeded %s",p);
	else { LOCK_TR1(") lock failed %s",p); }
	FREE(p);
#endif
	FREE(rrl);
	SYS_IMPORT;
	return rl.res;
}

static void reply_init_lock(vnn_t from,long l,long i)
{
	((LOCK)l)->LOCK_header=(FOB)i;
}
static void request_init_lock(vnn_t from,long l)
{
	LOCK_INFO *i;
	CALLOC(i,sizeof(LOCK_INFO),1);
	i->lck=RECVFOB((FOB)l,from);
	am_reply_2(from,reply_init_lock,l,(long)SENDFOB(i,from));
}
void p_init_lock_header(LOCK l)
{
	/* no need to do any locking, as this function is called
	 * before l can possibly be known to any other thread
	 * On the other hand, primary cannot be called on this
	 * lock, as it has not yet been defined.
	 */
	l->LOCK_header=NULL;
	am_request_1(lock_manager_cluster,request_init_lock,(long)l);
	am_wait_for(l->LOCK_header!=NULL);
}

static void request_endlock(vnn_t from,LOCK_PLATE n,long sem)
{
	DEBUG1("killing PLATE %p",n);
	/*
	 * There is a potential race condition between this code and process_end_plates.
	 * process_end_plates takes care of this by making a local copy of n->dead
	 */
	n->dead=ALL_DEAD;
	am_reply_1(from,r_sema_signal,sem);
	LCK_UNLOCK(wakeup_manager);
}

void r_endlock(void *vex)
{
#ifdef LOCK_TRACE
	char *p;
#endif
	EXCEPT_LOCK_ELEMENT(1) *ex=vex;
	sema_t sem;
/*	if(ex->plate==0) return; */
	SYS_EXPORT;
#ifdef LOCK_TRACE
	p=print_locks(ex,-1);
	LOCK_TR1("( endlock %s",p);
#endif
	if(ex->flags&LOCK_LOCAL) {
		if(!(ex->flags&LOCK_UNLOCKED)) {
			int rl;
			ex->flags|=LOCK_UNLOCKED;
			rl=lEXT_LOCK_fast_unlock(ex->locks[0],thr_ps_id());
			SYS_EXPORT;
			if(rl) am_request_0(lock_manager_cluster,awake_lock_manager);
#ifdef LOCK_TRACE
			LOCK_TR1(") endlock %p",(long)lck);
			FREE(p);
#endif
		} else SYS_EXPORT;
		return;
	}
	sem=SEMA_CREATE(0) ;
	am_request_2(lock_manager_cluster,(handler_2_t)request_endlock,(long)ex->plate,(long)sem);
#ifdef LOCK_TRACE
	LOCK_TR1(") endlock %s",p);
	FREE(p);
#endif
	SEMA_WAIT(sem);
	SEMA_DELETE(sem);
}


static void reply_unlock(vnn_t from,sema_t arg,long *res,long r)
{
	*res=r;
	SEMA_SIGNAL(arg);
}
static void request_unlock(vnn_t from,LOCK_PLATE p,FOB lck,long arg,long res)
{
	int i;
	for(i=0;i<p->cols;i++) {
		if(!(p->dead&(1<<i)) && p->lock[i]==lck) {
			p->dead|=1<<i;
			LCK_UNLOCK(wakeup_manager);
			am_reply_3(from,(handler_3_t)reply_unlock,arg,res,0);
			return;
		}
	}
	fprintf(stderr,"Sorry, but the program tried to make a bogus unlock\n");
	am_reply_3(from,(handler_3_t)reply_unlock,arg,res,-1);
	exit(-1);
}

void r_unlock(FOB lck)
{
	FOB f;
	long result;
	EXCEPT_LOCK_ELEMENT(1) *e;
	sema_t sem;

	e=((void *)((LOCAL_MEM)THR_LOCAL)->exception_stack);
	while(!e->c.type==EXCEPT_LOCK) e=(void *)e->c.prev_frame;
	LOCK_TR1("( unlocking %p",lck);

	if(e->flags&LOCK_LOCAL) {
		int rl;
		if(e->flags&LOCK_UNLOCKED) RFATAL("bogus unlock");
		e->flags|=LOCK_UNLOCKED;
		SYS_EXPORT;
		rl=lEXT_LOCK_fast_unlock(e->locks[0],thr_ps_id());
		if(rl) am_request_0(lock_manager_cluster,awake_lock_manager);
		LOCK_TR1(") unlocking %p",(long)lck);
		return;
	}
	sem=SEMA_CREATE(0);
	F_R_RATTR_NA(f,LOCK,lck,LOCK_header);
	SYS_EXPORT;
	am_request_4(lock_manager_cluster,(handler_4_t)request_unlock,(long)e->plate,(long)SENDFOBHOME(f),(long)sem,(long)&result);

	SEMA_WAIT(sem);
	if(result==-1) RFATAL("bogus unlock");
	SEMA_DELETE(sem);
	LOCK_TR1(") unlocking %p",(long)lck);
}


#ifdef DEADLOCK_DETECTION

void cDEBUG1INT(char *fmt, int val)
{
  char  buf[256];
  sprintf(buf, fmt, val);
  DEBUG0(buf);
}

#define PLATETABLESIZE 200

static LOCK_PLATE  firstPlateTable[PLATETABLESIZE];

static void init_first_lock_plate(void)
{
    int  i;

    for (i = 0; i < PLATETABLESIZE; ++i)
	firstPlateTable[i] = NULL;
}

static void  add_first_lock_plate(LOCK_PLATE  plate)
{
    int  hashValue = THR_HASH(plate->id) % PLATETABLESIZE;

    while (1) {
	if (!firstPlateTable[hashValue]) {
	    /* first entry in the hash table */
	    /* cDEBUG1INT("putting lock_plate into bucket #%d", hashValue); */
	    firstPlateTable[hashValue] = plate->first_branch;
	    return;
	}
	if (THR_SAME_ID(plate->id, firstPlateTable[hashValue]->id)) {
	    /* it's already in the table */
	    return;
	}
	/* someone else is in our spot */
	DEBUG0("collision in firstPlate hash table");
	++hashValue;
	hashValue %= PLATETABLESIZE;
    }
}

static LOCK_PLATE  find_threads_first_lock_plate(THREAD_ID id)
{
    int  hashValue;

    cDEBUG1INT("searching for thread %d first lock_plate", THR_HASH(id));
    hashValue = THR_HASH(id) % PLATETABLESIZE;
    /* cDEBUG1INT("hashValue = %d", hashValue); */
    while (firstPlateTable[hashValue]) {
	/* cDEBUG1INT("comparing to thread %d's lock_plate", THR_HASH(firstPlateTable[hashValue]->id)); */
	if (THR_SAME_ID(id, firstPlateTable[hashValue]->id)) {
	    return firstPlateTable[hashValue];
	}
	++hashValue;
	hashValue %= PLATETABLESIZE;
    }
    cDEBUG1INT("unable to find thread %d's first lock_plate", THR_HASH(id));
    return NULL;
}

static int  check_thread(THREAD_ID id);

/* A return of 0 means that this branch is stuck, no progress can be
   made, it's a deadlock.  A return of 1 means that there is some way
   of getting out of a deadlock. */
static int  check_branch(LOCK_PLATE  waiting_plate)
{
    LOCK_INFO  *lock_info;
    int i, j;
    struct ARRAYTHREAD_ID  {
      OB_HEADER header;
      int asize;
      THREAD_ID arr_part[1];
    } *wait_for;

    if (waiting_plate->dead == ALL_DEAD)
      return 1; /* we need to ignore dead plates */
    /* go through all desired locks and make sure that waiting on them */
    /* won't cause a deadlock */
    for (i = 0; i < waiting_plate->cols; ++i) {
        lock_info = waiting_plate->lock[i];
        wait_for = (struct ARRAYTHREAD_ID *)
	  lEXT_LOCK_wait_for(lock_info->lck, waiting_plate->id);
	cDEBUG1INT("check_branch waiting_plate = %p", (int) waiting_plate);
	cDEBUG1INT("wait_for = %p", (int) wait_for);
         /* are we waiting on anyone? wait_for is NULL when we arn't */
        if (wait_for) {
            for (j = 0; j < wait_for->asize; ++j) {
	      cDEBUG1INT("wait_for->arr_part[%d]", j);
                if (check_thread(wait_for->arr_part[j]) == 0)
                    return 0; /* if this thread is caught in a
                                 deadlock, then we'll never get a lock
                                 we want */
            }
        }
    }
    /* none of the branches will deadlock */
    return 1;
}

/* A return of 0 means that this thread is stuck, no progress can be
   made, it's a deadlock.  A return of 1 means that there is some way
   of getting out of a deadlock. */
static int  check_thread(THREAD_ID id)
{
    /* we need a way of mapping thread_id's into LOCK_PLATEs */
    LOCK_PLATE  waiting_plate = find_threads_first_lock_plate(id);
    cDEBUG1INT("check_thread: id = %d", THR_HASH(id));
    cDEBUG1INT("check_thread: waiting_plate = %p", (int) waiting_plate);
    if (!waiting_plate) {
	DEBUG0("waiting_plate is NULL");
	return 1;
    }
    if (waiting_plate->dead == ALL_DEAD) {
	DEBUG0("waiting_plate is dead");
	return 1;
    }
    if (waiting_plate->mark) {
	cDEBUG1INT("this plate %p is marked", (int) waiting_plate);
        return 0; /* we've been here before, we must be going in circles */
    }
    waiting_plate->mark = 1;
    /* the branches are disjunctive, we use a logical or; that means
       that if any of the branches will not deadlock, we are not in a
       deadlock situation */
    for (; waiting_plate; waiting_plate = waiting_plate->next_branch) {
        if (check_branch(waiting_plate) == 1)
            return 1; /* one of them won't deadlock */
    }
    /* all of them will deadlock; therefore, it really is a deadlock */
    cDEBUG1INT("all of the branches of thread %d will deadlock", THR_HASH(id));
    return 0;
}

static int  detect_deadlock(THREAD_ID id)
{
    int  ret = 0;
    LOCK_PLATE  plate;

    /* LCK_LOCK(top_plate); */

    DEBUG0("starting deadlock detect");
    init_first_lock_plate();
    /* clear the plates to prepare for the traversal */
    for (plate = top_plate.next; plate; plate = plate->next) {
	if(plate->dead!=ALL_DEAD) {
		plate->mark = 0;
		plate->first_branch->mark = 0;
		add_first_lock_plate(plate);
	}
    }

    ret = !check_thread(id);

    /* LCK_UNLOCK(top_plate); */
    cDEBUG1INT("ending deadlock detect, result=%d", ret);
#ifdef LOCK_DEBUG
    if (ret)
      RFATAL("deadlock detected\n");
#endif
    return ret;
}

#endif  /* DEADLOCK_DETECTION */

/*
 * call stubs for member function of locks
 */
extern void *EXT_LOCK_primary(LOCK);
extern int  EXT_LOCK_reservable(LOCK,THREAD_ID*);
extern int  EXT_LOCK_fast_try(LOCK,THREAD_ID*);
extern int  EXT_LOCK_fast_lock(LOCK,THREAD_ID*);
extern int  EXT_LOCK_fast_unlock(LOCK,THREAD_ID*);
extern void EXT_LOCK_reserve(LOCK,THREAD_ID*);
extern void EXT_LOCK_free(LOCK,THREAD_ID*);
extern void EXT_LOCK_request_reservation(LOCK,THREAD_ID*);
extern void EXT_LOCK_cancel_reservation(LOCK,THREAD_ID*);
extern void *EXT_LOCK_wait_for(LOCK,THREAD_ID*);
extern void *EXT_LOCK_combinations(LOCK);
extern void EXT_LOCK_combined_with(LOCK,FOB,FOB);
#define TID1(t)  (*((long *)&(t)))
#define TID2(t)  (*(((long *)&(t))+1))
#define SETTID(t,t1,t2) *(long *)&(t)=t1;*(((long *)&(t))+1)=t2
typedef struct value_lock {
	long value;
	DECL_L_LOCK(lo);
} VALUE_LOCK;
typedef struct np_value_lock {
	long value;
	NP_DECL_L_LOCK(lo);
} NP_VALUE_LOCK;
#define NP_VALUE(a)  (((NP_VALUE_LOCK*)(a))->value)
#define VALUE(a)  (((VALUE_LOCK*)(a))->value)

static void store_value(vnn_t from,long a,long b)
{
	*(void **)a=(void *)b;
}
static void store_value_unlock(vnn_t from,long a,long b,long l)
{
	*(void **)a=(void *)b;
	LOCK_L_UNLOCK((LCK_T)l);
}
static void store_value_lock_unlock(vnn_t from,long a,long b)
{
	VALUE(a)=b;
	LOCK_L_UNLOCK(((LCK(a))));
}
static void store_value_np_lock_unlock(vnn_t from,long a,long b)
{
	NP_VALUE(a)=b;
	NP_LOCK_L_UNLOCK(((NP_LCK(a))));
}
static void qEXT_LOCK_primary(vnn_t from,void *l,void *p,long lo)
{
	void *r;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	r=EXT_LOCK_primary(l);
	am_request_3(from,store_value_unlock,(long)p,(long)SENDFOB(r,from),lo);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static void *lEXT_LOCK_primary(LOCK l)
{
	int w;
	volatile void *p=(void *)1;
	DECL_L_LOCK(lo);
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		return EXT_LOCK_primary(l);
	INIT_L_LOCK(lo);
	/* printf("enter remote primary\n"); */
	thr_create_3(w,(handler_3_t)qEXT_LOCK_primary,(long)SENDFOBHOME(l),(long)&p,LOCK_PASS(lo));
	LOCK_L_LOCK(lo);
	/* printf("exit remote primary\n"); */
	return (void *)p;
}
static void qEXT_LOCK_reservable(vnn_t from,LOCK l,long t1,long t2,long rr)
{
	int r;
	THREAD_ID t;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	SETTID(t,t1,t2);
	r=EXT_LOCK_reservable(l,&t);
	am_request_2(from,store_value_lock_unlock,rr,r);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static int lEXT_LOCK_reservable(LOCK_PLATE plate,LOCK_INFO *li,unsigned long seq,THREAD_ID tid)
{
	int w;
	volatile VALUE_LOCK rlo;
	char p[30];
	if(li->res_seq==seq && li->res_plate==plate && THR_SAME_ID(li->res_tid,tid)) return li->reservable;
	li->res_seq=seq;
	li->res_plate=plate;
	li->res_tid=tid;
	/* printf("starting reservable for %s\n",thr_print_id(tid,p)); */
#ifndef LOCAL_EXECUTION
	if((w=WHERE(li->lck))==HERE) 
#endif
		return (li->reservable=EXT_LOCK_reservable(li->lck,&tid));
	INIT_L_LOCK(rlo.lo);
	/* printf("enter remote reservable\n"); */
	thr_create_4(w,(handler_4_t)qEXT_LOCK_reservable,(long)SENDFOBHOME(li->lck),TID1(tid),TID2(tid),(long)&rlo);
	LOCK_L_LOCK(rlo.lo);
	/* printf("exit remote reservable\n"); */
	/* printf("reservable done\n"); */
	return li->reservable=rlo.value;
}
static void qEXT_LOCK_fast_lock(vnn_t from,LOCK l,long t1,long t2,long rr)
{
	int r;
	THREAD_ID t;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	SETTID(t,t1,t2);
	r=EXT_LOCK_fast_lock(l,&t);
	am_request_2(from,store_value_np_lock_unlock,rr,r);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static int lEXT_LOCK_fast_lock(LOCK l,THREAD_ID tid)
{
	int w;
	NP_VALUE_LOCK rlo;
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		return EXT_LOCK_fast_lock(l,&tid);
	NP_INIT_L_LOCK(rlo.lo);
	/* printf("enter remote fast lock\n"); */
	thr_create_4(w,(handler_4_t)qEXT_LOCK_fast_lock,(long)SENDFOBHOME(l),TID1(tid),TID2(tid),(long)&rlo);
	NP_LOCK_L_LOCK(rlo.lo);
	/* printf("exit remote fast lock\n"); */
	return rlo.value;
}
static void qEXT_LOCK_fast_try(vnn_t from,LOCK l,long t1,long t2,long rr)
{
	int r;
	THREAD_ID t;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	SETTID(t,t1,t2);
	r=EXT_LOCK_fast_try(l,&t);
	am_request_2(from,store_value_lock_unlock,rr,r);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static int lEXT_LOCK_fast_try(LOCK l,THREAD_ID tid)
{
	int w;
	VALUE_LOCK rlo;
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		return EXT_LOCK_fast_try(l,&tid);
	INIT_L_LOCK(rlo.lo);
	/* printf("enter remote fast try %p\n",rlo.lo); */
	thr_create_4(w,(handler_4_t)qEXT_LOCK_fast_try,(long)SENDFOBHOME(l),TID1(tid),TID2(tid),(long)&rlo);
	LOCK_L_LOCK(rlo.lo);
	/* printf("exit remote fast try\n"); */
	return rlo.value;
}
static void qEXT_LOCK_fast_unlock(vnn_t from,LOCK l,long t1,long t2,long rr)
{
	int r;
	THREAD_ID t;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	SETTID(t,t1,t2);
	r=EXT_LOCK_fast_unlock(l,&t);
	am_request_2(from,store_value_lock_unlock,rr,r);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static int lEXT_LOCK_fast_unlock(LOCK l,THREAD_ID tid)
{
	int w;
	VALUE_LOCK rlo;
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		return EXT_LOCK_fast_unlock(l,&tid);
	INIT_L_LOCK(rlo.lo);
	/* printf("enter remote fast unlock\n"); */
	thr_create_4(w,(handler_4_t)qEXT_LOCK_fast_unlock,(long)SENDFOBHOME(l),TID1(tid),TID2(tid),(long)&rlo);
	LOCK_L_LOCK(rlo.lo);
	/* printf("exit remote fast unlock\n"); */
	return rlo.value;
}
static void qEXT_LOCK_reserve(vnn_t from,LOCK l,long t1,long t2,long rr)
{
	THREAD_ID t;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	SETTID(t,t1,t2);
	EXT_LOCK_reserve(l,&t);
	am_request_1(from,r_lck_unlock,rr);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static void lEXT_LOCK_reserve(LOCK l,THREAD_ID tid)
{
	int w;
	DECL_L_LOCK(lo);
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) { EXT_LOCK_reserve(l,&tid); 
#endif
		return; }
	INIT_L_LOCK(lo);
	/* printf("enter remote reserve\n"); */
	thr_create_4(w,(handler_4_t)qEXT_LOCK_reserve,(long)SENDFOBHOME(l),TID1(tid),TID2(tid),LOCK_PASS(lo));
	LOCK_L_LOCK(lo);
	/* printf("exit remote reserve\n"); */
}
static void qEXT_LOCK_free(vnn_t from,LOCK l,long t1,long t2,long rr)
{
	THREAD_ID t;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	SETTID(t,t1,t2);
	EXT_LOCK_free(l,&t);
	am_request_1(from,r_lck_unlock,rr);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static void lEXT_LOCK_free(LOCK l,THREAD_ID tid)
{
	int w;
	DECL_L_LOCK(lo);
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		return EXT_LOCK_free(l,&tid);
	INIT_L_LOCK(lo);
	/* printf("enter remote free\n"); */
	thr_create_4(w,(handler_4_t)qEXT_LOCK_free,(long)SENDFOBHOME(l),TID1(tid),TID2(tid),LOCK_PASS(lo));
	LOCK_L_LOCK(lo);
	/* printf("exit remote free\n"); */
}
static void qEXT_LOCK_request_reservation(vnn_t from,LOCK l,long t1,long t2,long rr)
{
	THREAD_ID t;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	SETTID(t,t1,t2);
	EXT_LOCK_request_reservation(l,&t);
	am_request_1(from,r_lck_unlock,rr);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static void lEXT_LOCK_request_reservation(LOCK l,THREAD_ID tid)
{
	int w;
	DECL_L_LOCK(lo);
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		return EXT_LOCK_request_reservation(l,&tid);
	INIT_L_LOCK(lo);
	/* printf("enter remote request_reservation\n"); */
	thr_create_4(w,(handler_4_t)qEXT_LOCK_request_reservation,(long)SENDFOBHOME(l),TID1(tid),TID2(tid),LOCK_PASS(lo));
	LOCK_L_LOCK(lo);
	/* printf("exit remote request_reservation\n"); */
}
static void qEXT_LOCK_cancel_reservation(vnn_t from,LOCK l,long t1,long t2,long rr)
{
	THREAD_ID t;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	SETTID(t,t1,t2);
	/* printf("executing cancel request on cluster %d\n",HERE); */
	EXT_LOCK_cancel_reservation(l,&t);
	am_request_1(from,r_lck_unlock,rr);
	/* printf("executing cancel request on cluster %d done\n",HERE); */
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static void lEXT_LOCK_cancel_reservation(LOCK l,THREAD_ID tid)
{
	int w;
	volatile int r=0;
	DECL_L_LOCK(lo);
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		return EXT_LOCK_cancel_reservation(l,&tid);
	/* printf("sending cancel request to cluster %d\n",w); */
	INIT_L_LOCK(lo);
	thr_create_4(w,(handler_4_t)qEXT_LOCK_cancel_reservation,(long)SENDFOBHOME(l),TID1(tid),TID2(tid),LOCK_PASS(lo));
	/* printf("lock manager waits for %p\n",lo); */
	LOCK_L_LOCK(lo);
	/* printf("cancel request to cluster %d done\n",w); */
}
#ifdef DEADLOCK_DETECTION
static void qEXT_LOCK_wait_for(vnn_t from,LOCK l,long t1,long t2,long rr)
{
	void *r;
	THREAD_ID t;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	SETTID(t,t1,t2);
	r=EXT_LOCK_wait_for(l,&t);
	am_request_3(from,store_value_lock_unlock,(long)SENDFOB(r,from));
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static void *lEXT_LOCK_wait_for(LOCK l,THREAD_ID tid)
{
	int w;
	VALUE_LOCK  rlo;
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		return EXT_LOCK_wait_for(l,&tid);
	INIT_L_LOCK(rlo.lo);
	/* printf("enter wait for\n"); */
	thr_create_4(w,(handler_4_t)qEXT_LOCK_wait_for,(long)SENDFOBHOME(l),TID1(tid),TID2(tid),(long)&rlo);
	LOCK_L_LOCK(rlo.lo);
	/* printf("exit wait for\n"); */
	return (void *)rlo.value;
}
#endif
static void qEXT_LOCK_combinations(vnn_t from,void *l,long p,long lo)
{
	void *r;
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	r=EXT_LOCK_combinations(l);
	am_request_3(from,store_value_unlock,p,(long)SENDFOB(r,from),lo);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static void *lEXT_LOCK_combinations(LOCK l)
{
	int w;
	volatile void *p=(void *)1;
	DECL_L_LOCK(lo);
#ifndef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		return EXT_LOCK_combinations(l);
	INIT_L_LOCK(lo);
	/* printf("enter remote combinations\n"); */
	thr_create_3(w,(handler_3_t)qEXT_LOCK_combinations,(long)SENDFOBHOME(l),(long)&p,LOCK_PASS(lo));
	LOCK_L_LOCK(lo);
	/* printf("exit remote combinations\n"); */
	return (void *)p;
}
static void qEXT_LOCK_combined_with(vnn_t from,LOCK l,FOB t1,FOB t2,long rr)
{
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	init_thread();
#else
	init_thread_memory();
#endif
	INIT_PROTECT_BEGIN
	EXT_LOCK_combined_with(l,RECVFOB(t1,from),RECVFOB(t2,from));
	am_request_1(from,r_lck_unlock,rr);
	INIT_PROTECT_END
#if defined(DEBUG) || defined(PSATHER_TRACE) || defined(PRINT_BACKTRACE)
	end_thread();
#else
	end_thread_memory();
#endif
}
static void lEXT_LOCK_combined_with(LOCK l,FOB f1,FOB f2)
{
	int w;
	DECL_L_LOCK(lo);
#ifdef LOCAL_EXECUTION
	if((w=WHERE(l))==HERE) 
#endif
		{ EXT_LOCK_combined_with(l,f1,f2); return; }
	INIT_L_LOCK(lo);
	/* printf("enter remote combined with\n"); */
	thr_create_4(w,(handler_4_t)qEXT_LOCK_combined_with,(long)SENDFOBHOME(l),(long)f1,(long)f2,LOCK_PASS(lo));
	LOCK_L_LOCK(lo);
	/* printf("exit remote combined with\n"); */
}


static void rll_lock_lck_remote(vnn_t from,LL_LOCK l,long r)
{
	LCK_LOCK(l->lck);
	am_request_2(from,store_value,r,-1);
}
void ll_lock_lck_remote(LL_LOCK l)
{
	int w;
	if((w=WHERE(l))==HERE) { LCK_LOCK(l->lck); return; }
	thr_create_2(w,(handler_2_t)rll_lock_lck_remote,(long)SENDFOBHOME(l),(long)&w);
	am_wait_for(w<0);
}
static void rll_lock_unlck_remote(vnn_t from,LL_LOCK l,long r)
{
	LCK_UNLOCK(l->lck);
	am_request_2(from,store_value,r,-1);
}
void ll_lock_unlck_remote(LL_LOCK l)
{
	int w;
	if((w=WHERE(l))==HERE) { LCK_UNLOCK(l->lck); return; }
	thr_create_2(w,(handler_2_t)rll_lock_unlck_remote,(long)SENDFOBHOME(l),(long)&w);
	am_wait_for(w<0);
}
static void rll_lock_try_remote(vnn_t from,LL_LOCK l,long r)
{
	int rr;
	rr=LCK_TRY(l->lck);
	am_request_2(from,store_value,r,rr);
}
int ll_lock_try_remote(LL_LOCK l)
{
	int w,r=-99;
	if((w=WHERE(l))==HERE) return LCK_TRY(l->lck);
	thr_create_2(w,(handler_2_t)rll_lock_try_remote,(long)SENDFOBHOME(l),(long)&r);
	am_wait_for(r!=-99);
	return r;
}


#ifdef SPINLOCK_LOCK
int SPINLOCK_TRY(SPINLOCK x) {
	register int r;
	TEST_AND_SET(&(x->lck),r);
	return !r;
}
static void rspinlock_lck_remote(vnn_t from,SPINLOCK l,long r)
{
	SPINLOCK_LCK(l);
	am_request_2(from,store_value,r,-1);
}
void spinlock_lck_remote(SPINLOCK l)
{
	int w;
	if((w=WHERE(l))==HERE) { SPINLOCK_LCK(l); return; }
	thr_create_2(w,(handler_2_t)rspinlock_lck_remote,(long)SENDFOBHOME(l),(long)&w);
	am_wait_for(w<0);
}
static void rspinlock_unlck_remote(vnn_t from,SPINLOCK l,long r)
{
	SPINLOCK_UNLCK(l);
	am_request_2(from,store_value,r,-1);
}
void spinlock_unlck_remote(SPINLOCK l)
{
	int w;
	if((w=WHERE(l))==HERE) { SPINLOCK_UNLCK(l); return; }
	thr_create_2(w,(handler_2_t)rspinlock_unlck_remote,(long)SENDFOBHOME(l),(long)&w);
	am_wait_for(w<0);
}
static void rspinlock_try_remote(vnn_t from,SPINLOCK l,long r)
{
	int rr;
	rr=SPINLOCK_TRY(l);
	am_request_2(from,store_value,r,rr);
}
int spinlock_try_remote(SPINLOCK l)
{
	int w,r=-99;
	if((w=WHERE(l))==HERE) return SPINLOCK_TRY(l);
	thr_create_2(w,(handler_2_t)rspinlock_try_remote,(long)SENDFOBHOME(l),(long)&r);
	am_wait_for(r!=-99);
	return r;
}
#endif
